from . import constants
import ipaddress
import socket
import struct
import math
import numpy as np


class P1255:
    def __init__(self):
        self.sock = None

    def connect(self, address, port=3000):
        """Connect to the P1255 oscilloscope at the specified address and port."""
        if not isinstance(address, ipaddress.IPv4Address):
            try:
                address = ipaddress.IPv4Address(address)
            except ipaddress.AddressValueError:
                raise ValueError(f"Not a valid IPv4 address: {str(address)}")

        # Validate port
        if not isinstance(port, int) or not (0 < port < 65536):
            raise ValueError(f"Not a valid port number, must be in between 0 and 65534: {str(port)}")

        # Create a TCP/IPv4 Socket
        self.sock = socket.socket(
            socket.AF_INET,  # Address family: IPv4
            socket.SOCK_STREAM,  # Socket type: TCP
        )

        self.sock.settimeout(1)  # 1 second timeout
        # Connect to the client device
        try:
            self.sock.connect((str(address), port))
        except Exception as e:
            self.sock.close()
            self.sock = None
            raise e
        return True

    def capture(self):
        if self.sock is None:
            return None
        try:
            # Send command to start streaming of binary data
            self.sock.send(b"STARTBIN")
            self.sock.settimeout(1)  # 1 second timeout

            # First information that is sent is the length of the dataset
            read = self.sock.recv_into(payload := bytearray(2), 2)
            if read != 2:
                raise RuntimeError("Length of dataset is not valid")
            length = struct.unpack("<H", payload)[0] + constants.LEN_UNKNOWN

            buffer = bytearray(length)
            buffer[:2] = payload

            while read < length:
                n = self.sock.recv_into(memoryview(buffer)[read:], length - read)
                if n == 0:
                    raise ConnectionError("Socket connection lost during data capture.")
                read += n

            return Dataset(buffer)
        except (TimeoutError, ConnectionError) as e:
            raise ConnectionError(f"Socket error during capture: {e}")

    def disconnect(self):
        """Disconnect from the P1255 oscilloscope."""
        if self.sock:
            self.sock.close()
            self.sock = None

    def __del__(self):
        self.disconnect()


class Dataset:

    class Channel:
        def __init__(self, buffer: memoryview) -> None:
            self.buffer = buffer

            # Channel name
            self.name = str(buffer[:3], 'utf8')

            # Timescale information
            # How long is the timescale in which the total channel data was captured
            def calc_timescale(number):
                exp = math.floor(number / 3)
                mant = {0: 1, 1: 2, 2: 5}[number % 3]
                time_per_div = mant * (10 ** exp)
                return 15 * time_per_div * 1e-9  # times 15 divisions on the screen, convert from nanoseconds to seconds
            self.timescale = calc_timescale(self.buffer[constants.CHANNEL_TIMESCALE])

            # Voltage scaling information
            def calc_voltscale(number):
                number += 4
                exp = math.floor(number / 3) - 1  # dont know why -1
                mant = {0: 1, 1: 2, 2: 5}[number % 3]
                volts_per_div = mant * (10 ** exp)
                return volts_per_div * 1e-3  # convert from millivolts to volts
            self.voltscale = calc_voltscale(self.buffer[constants.CHANNEL_VOLTSCALE])

            # Voltage shift # Julian: I think this is the offset in 1/25 of a div.
            self.volts_offset = struct.unpack(
                '<l',
                self.buffer[constants.CHANNEL_OFFSET:constants.CHANNEL_OFFSET + 4]
            )[0]

            # Get the data points from the buffer
            # '<h' corresponds to little endian signed short, times the number of samples
            self.data = np.array([
                (x / 128) * 5 * self.voltscale for x in  # apply this transformation to all data points.
                # The (x / 128) * 5 transforms the data into the unit on the screen,
                # the self.voltscale factor scales it to volts.
                struct.unpack(
                    '<' + 'h' * ((len(self.buffer) - constants.BEGIN_CHANNEL_DATA) // 2),  # specify data format
                    self.buffer[constants.BEGIN_CHANNEL_DATA:]  # specify the slice of the dataset
                )
            ])

            self.data_divisions = self.data / self.voltscale + self.volts_offset / 25

    def __init__(self, buffer: bytearray) -> None:
        self._buffer = buffer
        self.channels = list()

        # The model name and serial number of the oscilloscope
        # starts at 0x16 and is 12 bytes long
        # first 5 digits are the model name, the rest is the serial number
        serial_raw = buffer[constants.BEGIN_SERIAL_STRING:constants.BEGIN_SERIAL_STRING + constants.LEN_SERIAL_STRING]
        self.model, self.serial = str(serial_raw[:5], 'utf8'), str(serial_raw[6:], 'utf8')

        # Number of channels in dataset = number of set bits in byte 0x35
        num_channels = buffer[constants.CHANNEL_BITMAP].bit_count()

        # Get the length of the dataset but
        # remove the 12 additional bits from the length of the dataset
        # Calculate the region of each channel
        channel_data_size = (len(buffer) - constants.LEN_HEADER) // num_channels

        for ch in range(num_channels):
            # Get a slice of the dataset and let the channel class do its work
            # The slices first have an offset (header) and then they are concatenated
            # Append this to the list of channels
            self.channels.append(
                Dataset.Channel(
                    memoryview(buffer)[constants.LEN_HEADER + ch * channel_data_size:constants.LEN_HEADER + (ch + 1) * channel_data_size]
                )
            )

    def save(self, filename, fmt='csv'):
        """Save the dataset to a file in the specified format."""
        if fmt == 'json':
            import json
            data = [{"name": ch.name, "timescale": ch.timescale, "data": ch.data.tolist()} for ch in self.channels]
            with open(filename, 'w') as f:
                json.dump(data, f)
        elif fmt == 'csv':
            import csv
            with open(filename, 'w', newline='') as f:
                writer = csv.writer(f)
                writer.writerow(['Time (s)'] + [f"{ch.name} (V)" for ch in self.channels])  # write header
                # Calculate timescale information
                time = np.linspace(start=(-1) * self.channels[0].timescale / 2, stop=self.channels[0].timescale / 2, num=len(self.channels[0].data), endpoint=True)
                # write the data with the time column
                write_data = [time] + [ch.data for ch in self.channels]
                writer.writerows(zip(*write_data))
        elif fmt == 'npz':
            data = {
                **{ch.name: ch.data for ch in self.channels},
                'time': np.linspace(start=(-1) * self.channels[0].timescale / 2, stop=self.channels[0].timescale / 2, num=len(self.channels[0].data), endpoint=True)
            }
            np.savez(filename, **data)  # save as .npz file
        else:
            raise ValueError("Unsupported format. Use 'csv', 'json' or 'npz'.")



#!/usr/bin/env python3


# Orientation within the dataset
BEGIN_SERIAL_STRING = 0x16
LEN_SERIAL_STRING = 12

CHANNEL_BITMAP = 0x35

LEN_UNKNOWN = 12  # The weird 12 bits that were figured out by trial and error
LEN_HEADER = 66

BEGIN_CHANNEL_DATA = 0x3b
CHANNEL_TIMESCALE = 0x1b
CHANNEL_VOLTSCALE = 0x23
CHANNEL_OFFSET = 0x1f


CONNECTION_HELP = """P1255 Connection Help:

    - Connect the Oscilloscope to a network via a LAN cable.
    - Press the "utility" button on the oscilloscope.
    - Press the "H1" button to access the possible menus.
    - Scroll down to "LAN Set" by rotating the "M" knob.
    - Press the "M" knob to enter the menu.
    - Press on the "H2" Button ("Set").
    - You can use the "F*" buttons and the "M" knob to adjust other settings.
    - Save the changes and restart to apply them.
        """


from PyQt5.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QFileDialog,
)
from PyQt5 import uic
from PyQt5.QtCore import QTimer
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.ticker import MultipleLocator
import matplotlib.pyplot as plt
import os
from p1255.p1255 import P1255
from p1255.constants import CONNECTION_HELP
import ipaddress
from PyQt5.QtWidgets import QMessageBox
from pathlib import Path
import yaml
import importlib.resources
import numpy as np


plt.style.use('dark_background')

ALIAS_FILE = Path().home() / ".p1255_ip_aliases.yaml"
COLORS = {
    "CH1": 'red',
    "CH2": 'yellow',
}


class PlotWidget(FigureCanvas):
    def __init__(self, parent=None):
        self.fig = Figure()
        super().__init__(self.fig)
        self.ax = self.fig.add_subplot(111)

    def update_plot(self, dataset, unit, mode):
        """Update the plot with data and unit

        Parameters
        ----------
        dataset : Dataset
            The dataset to plot.
        unit : str
            The unit to plot ('Voltage' or 'Divisions').
        mode : str
            The mode of the oscilloscope ('Normal', 'X: Ch1, Y: Ch2', 'X: Ch2, Y: Ch1').
        """
        if unit not in ('Voltage', 'Divisions'):
            raise ValueError("Unit must be 'Voltage' or 'Divisions'")
        if mode not in ('Normal', 'X: Ch1, Y: Ch2', 'X: Ch2, Y: Ch1'):
            raise ValueError("Mode must be 'Normal', 'X: Ch1, Y: Ch2', or 'X: Ch2, Y: Ch1'")
        self.ax.clear()
        if dataset:
            if mode == 'Normal':
                time = np.linspace(start=(-1) * dataset.channels[0].timescale / 2, stop=dataset.channels[0].timescale / 2, num=len(dataset.channels[0].data), endpoint=True)
                if time[-1] < 1e-3:
                    time *= 1e6
                    self.ax.set_xlabel('Time (µs)')
                elif time[-1] < 1:
                    time *= 1e3
                    self.ax.set_xlabel('Time (ms)')
                else:
                    self.ax.set_xlabel('Time (s)')
                for i, channel in enumerate(dataset.channels):
                    if unit == 'Voltage':
                        self.ax.plot(time, channel.data, label=channel.name, color=COLORS[channel.name])
                        self.ax.set_ylabel('Voltage (V)')
                        self.ax.relim()
                        self.ax.autoscale_view()
                    else:  # Divisions
                        self.ax.plot(time, channel.data_divisions, label=channel.name, color=COLORS[channel.name])
                        self.ax.yaxis.set_major_locator(MultipleLocator(1))
                        self.ax.set_ylabel('Divisions')
                        self.ax.set_ylim(-5, 5)
                self.ax.legend()
            else:  # XY Plot
                ch1 = dataset.channels[0]
                ch2 = dataset.channels[1]
                if unit == 'Voltage':
                    ch1 = ch1.data
                    ch2 = ch2.data
                elif unit == 'Divisions':
                    ch1 = ch1.data_divisions
                    ch2 = ch2.data_divisions
                    self.ax.yaxis.set_major_locator(MultipleLocator(1))
                    self.ax.xaxis.set_major_locator(MultipleLocator(1))
                    self.ax.set_ylim(-5, 5)
                    self.ax.set_xlim(-5, 5)
                if mode == 'X: Ch1, Y: Ch2':
                    self.ax.plot(ch1, ch2)
                    self.ax.set_xlabel(f'{dataset.channels[0].name} ({unit})')
                    self.ax.set_ylabel(f'{dataset.channels[1].name} ({unit})')
                else:  # Ch2/Ch1
                    self.ax.plot(ch2, ch1)
                    self.ax.set_xlabel(f'{dataset.channels[1].name} ({unit})')
                    self.ax.set_ylabel(f'{dataset.channels[0].name} ({unit})')

        else:
            self.ax.text(0.5, 0.5, 'No Data', horizontalalignment='center', verticalalignment='center')
        self.ax.grid(True, linestyle='--', alpha=0.5)
        self.draw()


class MainWindow(QWidget):
    def __init__(self, disable_aliases=False):
        super().__init__()
        with importlib.resources.path("p1255", "gui.ui") as ui_file:
            uic.loadUi(ui_file, self)

        self.disable_aliases = disable_aliases

        self.plot_widget = PlotWidget()
        layout = QVBoxLayout(self.plot_placeholder)
        layout.addWidget(self.plot_widget)
        self.timer = None
        self.saving_directory = os.getcwd()

        self.p1255 = P1255()
        self.current_dataset = None

        if Path(ALIAS_FILE).is_file() and not self.disable_aliases and self.aliases:
            self.use_alias = True
            with open(ALIAS_FILE, "r") as f:
                self.aliases = yaml.safe_load(f)
        else:
            self.use_alias = False

        if self.use_alias:
            self.connection_stack.setCurrentIndex(1)
            self.alias_combo.addItems(self.aliases.keys())
            self.alias_combo.currentIndexChanged.connect(self.connect_to_ip)
        else:
            self.connection_stack.setCurrentIndex(0)

        self.connect_button.clicked.connect(self.connect_to_ip)
        self.help_button.setFixedWidth(30)
        self.help_button.clicked.connect(self.show_help)
        self.run_button.clicked.connect(self.toggle_run)
        self.capture_button.clicked.connect(self.capture_single)
        self.save_button.clicked.connect(self.save_data)
        self.unit_combo.currentIndexChanged.connect(self.update_current)
        self.display_mode_combo.currentIndexChanged.connect(self.update_current)

        if self.use_alias:
            self.connect_to_ip()
        self.capture_single()

    def show_help(self):
        QMessageBox.information(self, "Help", CONNECTION_HELP)

    def connect_to_ip(self):
        if self.use_alias:
            alias = self.alias_combo.currentText()
            ip, port = self.aliases[alias]
        else:
            ip = self.ip_input.text()
            port = self.port_input.text()
        print(f"Connecting to {ip}:{port}...")
        try:
            self.p1255.connect(ipaddress.IPv4Address(ip), int(port))
        except Exception as e:
            QMessageBox.critical(self, "Connection Error", f"Failed to connect to the oscilloscope: {e}")
            return
        self.connect_button.setText("Connected")

    def disconnect(self):
        self.p1255.disconnect()
        self.connect_button.setText("Connect")

    def toggle_run(self, checked):
        self.run_button.setChecked(checked)  # this is in case the button gets unchecked programmatically
        if checked:
            self.run_button.setText("Stop")
            self.start_updating()
        else:
            self.run_button.setText("Run Continuously")
            self.stop_updating()

    def update_current(self):
        self.plot_widget.update_plot(self.current_dataset, self.unit_combo.currentText(), self.display_mode_combo.currentText())

    def start_updating(self):
        self.timer = QTimer()
        self.timer.timeout.connect(self.capture_single)
        self.timer.start(500)  # milliseconds

    def stop_updating(self):
        if self.timer:
            self.timer.stop()
            self.timer = None

    def capture_single(self):
        try:
            self.current_dataset = self.p1255.capture()
            self.plot_widget.update_plot(self.current_dataset, self.unit_combo.currentText(), self.display_mode_combo.currentText())
        except ConnectionError:
            QMessageBox.critical(self, "Connection Error", "Connection lost.")
            self.toggle_run(False)
            self.disconnect()
        except Exception as e:
            QMessageBox.critical(self, "Capture Error", f"Failed to capture data: {e}")
            self.toggle_run(False)
            self.disconnect()

    def save_data(self):
        if not self.current_dataset:
            print("No data to save.")
            return

        filename = QFileDialog.getSaveFileName(
            self, "Save Data", self.saving_directory, "CSV Files (*.csv);;JSON Files (*.json);;Numpy Files (*.npz)"
        )[0]
        if not filename:
            return

        ext = Path(filename).suffix.lower()
        fmt = ext.lstrip('.')
        if fmt in ('csv', 'json', 'npz'):
            self.current_dataset.save(filename, fmt=fmt)
